"""
Provide tool functions for working with Job Submission data
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, cast

from buzz import enforce_defined
from jobbergate_core.tools.sbatch import SubmissionHandler, inject_sbatch_params
from loguru import logger

from jobbergate_cli.config import settings
from jobbergate_cli.constants import FileType
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import JobbergateContext, JobSubmissionCreateRequestData, JobSubmissionResponse
from jobbergate_cli.subapps.job_scripts.tools import download_job_script_files


def _map_cluster_name(
    jg_ctx: JobbergateContext,
    base_cluster_name: str,
):
    """
    Injects the organization name into the cluster name for multi-tenancy mode.

    If the organization is undefined (multi-tenancy is disabled) or the cluster_name already includes the
    organization_id, use the base_cluster_name.
    """
    # Make static type checkers happy
    assert jg_ctx.persona is not None, "jg_ctx.persona is uninitialized"

    org_id = jg_ctx.persona.identity_data.organization_id
    if org_id is None or base_cluster_name.endswith(org_id):
        return base_cluster_name

    return f"{base_cluster_name}-{jg_ctx.persona.identity_data.organization_id}"


@dataclass
class JobSubmissionABC(ABC):
    """
    A dataclass representing a job submission for Jobbergate.

    :param: jg_ctx:                    The JobbergateContext. Used to retrieve the client for requests
                                       and the email of the submitting user
    :param: job_script_id:             The ``id`` of the Job Script to submit to Slurm
    :param: name:                      The name to attach to the Job Submission
    :param: description:               An optional description that may be added to the Job Submission
    :param: cluster_name:              An optional cluster_name for the cluster where the job should be executed,
                                       If left off, it will default to the DEFAULT_CLUSTER_NAME from the settings.
                                       If no default is set, an exception will be raised.
    :param: execution_directory:       An optional directory where the job should be executed. If provided as a
                                       relative path, it will be constructed as an absolute path relative to
                                       the current working directory.
    :param: download:                  A flag indicating whether the job script files should be downloaded to the.
    :param: sbatch_arguments:          An optional list of arguments to pass to inject into the job script.
    """

    jg_ctx: JobbergateContext
    job_script_id: int
    name: str
    execution_directory: Path | None = None
    cluster_name: str | None = None
    download: bool = False
    description: Optional[str] = None
    sbatch_arguments: Optional[list[str]] = None

    def __post_init__(self):
        """Post-init hook to ensure that the cluster_name and execution_directory are set correctly."""
        if not self.cluster_name:
            self.cluster_name = settings.DEFAULT_CLUSTER_NAME
        if self.cluster_name is None:
            raise ValueError("No cluster name supplied and no default exists. Cannot submit to an unknown cluster!")
        self.cluster_name = _map_cluster_name(self.jg_ctx, self.cluster_name)

        if self.execution_directory is None:
            self.execution_directory = Path.cwd()
        if not self.execution_directory.is_absolute():
            self.execution_directory = self.execution_directory.resolve()

    @abstractmethod
    def process_submission(self):
        """Process the job submission. This method should be overridden by subclasses."""
        pass

    def get_request_data(self) -> JobSubmissionCreateRequestData:
        """Return the data to be used in the POST request to the API."""
        return JobSubmissionCreateRequestData(
            name=self.name,
            description=self.description,
            job_script_id=self.job_script_id,
            cluster_name=self.cluster_name,
            execution_directory=self.execution_directory,
            sbatch_arguments=self.sbatch_arguments,
        )

    def make_post_request(self, job_submission_data: JobSubmissionCreateRequestData) -> JobSubmissionResponse:
        """Make the POST request to the API to create the job submission."""
        # Make static type checkers happy
        assert self.jg_ctx.client is not None, "jg_ctx.client is uninitialized"
        return cast(
            JobSubmissionResponse,
            make_request(
                self.jg_ctx.client,
                "/jobbergate/job-submissions",
                "POST",
                expected_status=201,
                abort_message="Couldn't create job submission",
                support=True,
                request_model=job_submission_data,
                response_model_cls=JobSubmissionResponse,
            ),
        )

    def run(self) -> JobSubmissionResponse:
        """Run all required steps to create the job submission."""
        self.process_submission()
        job_submission_data = self.get_request_data()
        return self.make_post_request(job_submission_data)


class RemoteJobSubmission(JobSubmissionABC):
    def process_submission(self):
        """Process the job submission in remote mode."""
        logger.info("Creating job submission in remote mode")
        if self.download:
            download_job_script_files(self.job_script_id, self.jg_ctx, self.execution_directory)


class OnsiteJobSubmission(JobSubmissionABC):
    def process_submission(self):
        """Process the job submission in on-site mode."""
        logger.info("Creating job submission in on-site mode")
        if settings.SBATCH_PATH is None:
            raise Abort(
                "SBATCH_PATH most be set for onsite submissions",
                subject="Configuration Error",
                support=True,
            )

        job_script_files = download_job_script_files(self.job_script_id, self.jg_ctx, self.execution_directory)

        entrypoint_file = [f for f in job_script_files if f.file_type == FileType.ENTRYPOINT]

        Abort.require_condition(
            len(entrypoint_file) == 1,
            f"There should be exactly one entrypoint file in the parent job script, got {len(entrypoint_file)}",
            raise_kwargs=dict(subject="Job Script Error"),
        )
        job_script_path = self.execution_directory / entrypoint_file[0].filename

        self.inject_sbatch_params(job_script_path)

        try:
            sbatch_handler = SubmissionHandler(
                sbatch_path=settings.SBATCH_PATH,
                submission_directory=self.execution_directory,
            )
            slurm_id = sbatch_handler.submit_job(job_script_path)
        except Exception as e:
            raise Abort(
                "Failed to submit job to Slurm",
                raise_kwargs=dict(
                    subject="Slurm Submission Error",
                    support=True,
                    log_message=f"On-site submission failed: {e:s}",
                ),
            ) from e
        self.slurm_job_id = slurm_id

    def inject_sbatch_params(self, job_script_path: Path):
        """Inject sbatch parameters into the job script."""
        if not self.sbatch_arguments:
            return
        job_script_content = job_script_path.read_text()
        job_script_content = inject_sbatch_params(
            job_script_content, self.sbatch_arguments, "Injected at submission time by Jobbergate CLI"
        )
        job_script_path.write_text(job_script_content)

    def get_request_data(self) -> JobSubmissionCreateRequestData:
        """Return the data to be used in the POST request to the API."""
        data = super().get_request_data()
        data.slurm_job_id = enforce_defined(
            getattr(self, "slurm_job_id", None),
            "Slurm job id not found, on-site submission failed",
            raise_exc_class=Abort,
            raise_kwargs=dict(
                subject="Slurm Submission Error",
                support=True,
            ),
        )
        return data


def job_submissions_factory(
    jg_ctx: JobbergateContext,
    job_script_id: int,
    name: str,
    execution_directory: Path | None = None,
    cluster_name: str | None = None,
    download: bool = False,
    description: Optional[str] = None,
    sbatch_arguments: Optional[list[str]] = None,
) -> JobSubmissionABC:
    """Job submission factory function. Returns the correct job submission class based on the current mode."""
    if settings.is_onsite_mode:
        return OnsiteJobSubmission(
            jg_ctx=jg_ctx,
            job_script_id=job_script_id,
            name=name,
            execution_directory=execution_directory,
            cluster_name=cluster_name,
            download=download,
            description=description,
            sbatch_arguments=sbatch_arguments,
        )
    return RemoteJobSubmission(
        jg_ctx=jg_ctx,
        job_script_id=job_script_id,
        name=name,
        execution_directory=execution_directory,
        cluster_name=cluster_name,
        download=download,
        description=description,
        sbatch_arguments=sbatch_arguments,
    )


def fetch_job_submission_data(
    jg_ctx: JobbergateContext,
    job_submission_id: int,
) -> JobSubmissionResponse:
    """
    Retrieve a job submission from the API by ``id``
    """
    # Make static type checkers happy
    assert jg_ctx.client is not None, "Client is uninitialized"

    return cast(
        JobSubmissionResponse,
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-submissions/{job_submission_id}",
            "GET",
            expected_status=200,
            abort_message=f"Couldn't retrieve job submission {job_submission_id} from API",
            support=True,
            response_model_cls=JobSubmissionResponse,
        ),
    )



"""
Provide a ``typer`` app that can interact with Job Submission data in a cruddy manner.
"""

from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Optional

import typer

from jobbergate_cli.constants import SortOrder
from jobbergate_cli.exceptions import Abort, handle_abort
from jobbergate_cli.render import StyleMapper, render_single_result, terminal_message
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import JobbergateContext, JobSubmissionResponse
from jobbergate_cli.subapps.job_submissions.tools import fetch_job_submission_data, job_submissions_factory
from jobbergate_cli.subapps.pagination import handle_pagination


# move hidden field logic to the API
HIDDEN_FIELDS = [
    "created_at",
    "execution_directory",
    "is_archived",
    "job_script",
    "report_message",
    "sbatch_arguments",
    "slurm_job_info",
    "updated_at",
]


style_mapper = StyleMapper(job_submission_id="green", name="cyan", slurm_job_id="dark_orange")


app = typer.Typer(help="Commands to interact with job submissions")


@app.command()
@handle_abort
def create(
    ctx: typer.Context,
    name: str = typer.Option(
        ...,
        "--name",
        "-n",
        help="The name of the job submission to create",
    ),
    description: Optional[str] = typer.Option(
        None,
        help="A helpful description of the job submission",
    ),
    job_script_id: int = typer.Option(
        ...,
        "--job-script-id",
        "-i",
        help="The id of the job_script from which to create the job submission",
    ),
    cluster_name: str = typer.Option(
        None,
        help="The name of the cluster where the job should be submitted (i.g. 'nash-staging')",
    ),
    execution_directory: Optional[Path] = typer.Option(
        None,
        help=dedent(
            """
            The path on the cluster where the job script should be executed.
            If provided as a relative path, it will be converted as an absolute path from your current
            working directory. If you use "~" to denote your home directory, the path will be expanded to an
            absolute path for your home directory on *this* machine.
            """
        ).strip(),
    ),
    sbatch_arguments: Optional[list[str]] = typer.Option(
        None,
        "--sbatch-arguments",
        "-s",
        help=dedent(
            """
            Additional arguments to pass as sbatch directives. These should be provided as a list of strings.
            See more details at: https://slurm.schedmd.com/sbatch.html
            """
        ).strip(),
    ),
    download: bool = typer.Option(
        False,
        help="Download the job script files to the current working directory",
    ),
):
    """
    Create a new job submission.
    """
    jg_ctx: JobbergateContext = ctx.obj

    try:
        submissions_handler = job_submissions_factory(
            jg_ctx,
            job_script_id,
            name,
            description=description,
            execution_directory=execution_directory,
            cluster_name=cluster_name,
            sbatch_arguments=sbatch_arguments,
            download=download,
        )
        result = submissions_handler.run()
    except Exception as err:
        raise Abort(
            "Failed to create the job submission",
            subject="Job submission failed",
            support=True,
            log_message=f"There was an issue submitting the job from job_script_id={job_script_id}",
            original_error=err,
        )

    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Created Job Submission",
    )


@app.command("list")
@handle_abort
def list_all(
    ctx: typer.Context,
    show_all: bool = typer.Option(
        False,
        "--all",
        help="Show all job submissions, even the ones owned by others",
    ),
    search: Optional[str] = typer.Option(None, help="Apply a search term to results"),
    sort_order: SortOrder = typer.Option(SortOrder.DESCENDING, help="Specify sort order"),
    sort_field: Optional[str] = typer.Option("id", help="The field by which results should be sorted"),
    from_job_script_id: Optional[int] = typer.Option(
        None,
        help="Filter job-submissions by the job-script-id they were created from.",
    ),
):
    """
    Show available job submissions.
    """
    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx is not None, "JobbergateContext is uninitialized"
    assert jg_ctx.client is not None, "Client is uninitialized"
    assert jg_ctx.persona is not None, "Persona is uninitialized"

    params: Dict[str, Any] = dict(user_only=not show_all)
    if search is not None:
        params["search"] = search
    if sort_order is not SortOrder.UNSORTED:
        params["sort_ascending"] = SortOrder is SortOrder.ASCENDING
    if sort_field is not None:
        params["sort_field"] = sort_field
    if from_job_script_id is not None:
        params["from_job_script_id"] = from_job_script_id

    value_mappers = None
    organization_id = jg_ctx.persona.identity_data.organization_id
    if organization_id is not None:
        value_mappers = dict(cluster_name=lambda cn: cn.replace(f"-{organization_id}", ""))

    handle_pagination(
        jg_ctx=jg_ctx,
        url_path="/jobbergate/job-submissions",
        abort_message="Couldn't retrieve job submissions list from API",
        params=params,
        title="Job Submission List",
        style_mapper=style_mapper,
        hidden_fields=HIDDEN_FIELDS,
        nested_response_model_cls=JobSubmissionResponse,
        value_mappers=value_mappers,
    )


@app.command()
@handle_abort
def get_one(
    ctx: typer.Context,
    id: int = typer.Option(..., "--id", "-i", help="The specific id of the job submission."),
):
    """
    Get a single job submission by id
    """
    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx is not None, "JobbergateContext is uninitialized"
    assert jg_ctx.persona is not None, "Persona is uninitialized"

    value_mappers = None
    organization_id = jg_ctx.persona.identity_data.organization_id
    if organization_id is not None:
        value_mappers = dict(cluster_name=lambda cn: cn.replace(f"-{organization_id}", ""))

    result = fetch_job_submission_data(jg_ctx, id)
    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Job Submission",
        value_mappers=value_mappers,
    )


# NOTE: job submissions update is not added because it was an effective noop on the former implementation


@app.command()
@handle_abort
def delete(
    ctx: typer.Context,
    id: int = typer.Option(
        ...,
        "--id",
        "-i",
        help="The id of the job submission to delete",
    ),
):
    """
    Delete an existing job submission.
    """
    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx.client is not None, "Client is uninitialized"

    make_request(
        jg_ctx.client,
        f"/jobbergate/job-submissions/{id}",
        "DELETE",
        expected_status=204,
        abort_message="Request to delete job submission was not accepted by the API",
        support=True,
    )
    terminal_message(
        "The job submission was successfully deleted.",
        subject="Job submission delete succeeded",
    )



"""
Configuration file, sets all the necessary environment variables.
Can load configuration from a dotenv file if supplied.
"""

from pathlib import Path
from sys import exit
from typing import Optional

from pydantic import AnyHttpUrl, BaseSettings, Field, ValidationError, root_validator

from jobbergate_cli import constants
from jobbergate_cli.constants import OV_CONTACT
from jobbergate_cli.render import terminal_message
from jobbergate_cli.text_tools import conjoin


class Settings(BaseSettings):
    """
    Provide a ``pydantic`` settings model to hold configuration values loaded from the environment.
    """

    JOBBERGATE_CACHE_DIR: Path = Field(Path.home() / ".local/share/jobbergate3")

    ARMADA_API_BASE: AnyHttpUrl = Field("https://armada-k8s.staging.omnivector.solutions")

    SBATCH_PATH: Optional[Path]

    # enable http tracing
    JOBBERGATE_DEBUG: bool = Field(False)
    JOBBERGATE_REQUESTS_TIMEOUT: Optional[int] = 15

    # Setry's configuration
    SENTRY_DSN: Optional[str]
    SENTRY_TRACE_RATE: float = Field(1.0, gt=0.0, le=1.0)
    SENTRY_ENV: str = "LOCAL"

    # Default job submission cluster
    DEFAULT_CLUSTER_NAME: Optional[str]

    # How long it will use cached cluster lists before fetching them again
    JOBBERGATE_CLUSTER_CACHE_LIFETIME: int = 60 * 5

    # Settings for log uploads
    JOBBERGATE_AWS_ACCESS_KEY_ID: Optional[str]
    JOBBERGATE_AWS_SECRET_ACCESS_KEY: Optional[str]
    JOBBERGATE_S3_LOG_BUCKET: str = Field("jobbergate-cli-logs")

    # Computed values. Listed as Optional, but will *always* be set (or overridden) based on other values
    JOBBERGATE_APPLICATION_MODULE_PATH: Optional[Path]
    JOBBERGATE_APPLICATION_CONFIG_PATH: Optional[Path]
    JOBBERGATE_LOG_PATH: Optional[Path]
    JOBBERGATE_USER_TOKEN_DIR: Optional[Path]
    JOBBERGATE_ACCESS_TOKEN_PATH: Optional[Path]
    JOBBERGATE_REFRESH_TOKEN_PATH: Optional[Path]

    JOBBERGATE_CLUSTER_LIST_PATH: Optional[Path]

    # Compatibility mode: If True, add commands as they appear in the legacy app
    JOBBERGATE_COMPATIBILITY_MODE: Optional[bool] = False
    JOBBERGATE_LEGACY_NAME_CONVENTION: Optional[bool] = False

    # Auth0 config for machine-to-machine security
    OIDC_DOMAIN: str
    OIDC_AUDIENCE: str
    OIDC_CLIENT_ID: str
    OIDC_USE_HTTPS: bool = True
    OIDC_MAX_POLL_TIME: int = 5 * 60  # 5 Minutes

    # Enable multi-tenancy to fix cluster name mapping by client_id
    MULTI_TENANCY_ENABLED: bool = False

    @root_validator(skip_on_failure=True)
    def compute_extra_settings(cls, values):
        """
        Compute settings values that are based on other settings values.
        """
        values["JOBBERGATE_CACHE_DIR"] = Path(values["JOBBERGATE_CACHE_DIR"]).expanduser().resolve()
        cache_dir = values["JOBBERGATE_CACHE_DIR"]
        cache_dir.mkdir(exist_ok=True, parents=True)

        values["JOBBERGATE_APPLICATION_MODULE_PATH"] = cache_dir / constants.JOBBERGATE_APPLICATION_MODULE_FILE_NAME
        values["JOBBERGATE_APPLICATION_CONFIG_PATH"] = cache_dir / constants.JOBBERGATE_APPLICATION_CONFIG_FILE_NAME

        log_dir = cache_dir / "logs"
        log_dir.mkdir(exist_ok=True, parents=True)
        values["JOBBERGATE_LOG_PATH"] = log_dir / "jobbergate-cli.log"

        token_dir = cache_dir / "token"
        token_dir.mkdir(exist_ok=True, parents=True)
        values["JOBBERGATE_USER_TOKEN_DIR"] = token_dir
        values["JOBBERGATE_ACCESS_TOKEN_PATH"] = token_dir / "access.token"
        values["JOBBERGATE_REFRESH_TOKEN_PATH"] = token_dir / "refresh.token"

        values["JOBBERGATE_CLUSTER_LIST_PATH"] = cache_dir / "clusters.json"

        return values

    @property
    def is_onsite_mode(self) -> bool:
        """Check if the SBATCH_PATH is set, indicating that the CLI is running in on-site mode."""
        return self.SBATCH_PATH is not None

    class Config:
        """
        Customize behavior of the Settings class. Especially, enable the use of dotenv to load settings from a ``.env``
        file instead of the environment.
        """

        if constants.JOBBERGATE_DEFAULT_DOTENV_PATH.is_file():
            env_file = constants.JOBBERGATE_DEFAULT_DOTENV_PATH
        else:
            env_file = Path(".env")


def build_settings(*args, **kwargs):
    """
    Return a Setting object and handle ValidationError with a message to the user.
    """
    try:
        return Settings(*args, **kwargs)
    except ValidationError as err:
        terminal_message(
            conjoin(
                "A configuration error was detected.",
                "",
                f"[yellow]Please contact [bold]{OV_CONTACT}[/bold] for support and trouble-shooting[/yellow]",
                "",
                "Details:",
                "",
                f"[red]{err}[/red]",
            ),
            subject="Configuration Error",
        )
        exit(1)


settings = build_settings()



"""
Provide compatibility to the previous version of Jobbergate CLI for users who have automation or are
familiar with the old commands
"""

import typer

from jobbergate_cli.subapps.applications.app import clone as clone_application
from jobbergate_cli.subapps.applications.app import create as create_application
from jobbergate_cli.subapps.applications.app import delete as delete_application
from jobbergate_cli.subapps.applications.app import download_files as download_files_application
from jobbergate_cli.subapps.applications.app import get_one as get_application
from jobbergate_cli.subapps.applications.app import list_all as list_applications
from jobbergate_cli.subapps.applications.app import update as update_application
from jobbergate_cli.subapps.job_scripts.app import clone as clone_job_script
from jobbergate_cli.subapps.job_scripts.app import delete as delete_job_script
from jobbergate_cli.subapps.job_scripts.app import download_files as download_files_job_script
from jobbergate_cli.subapps.job_scripts.app import get_one as get_job_script
from jobbergate_cli.subapps.job_scripts.app import list_all as list_job_scripts
from jobbergate_cli.subapps.job_scripts.app import render as render_job_script
from jobbergate_cli.subapps.job_scripts.app import render_locally as render_job_script_locally
from jobbergate_cli.subapps.job_scripts.app import show_files as show_files
from jobbergate_cli.subapps.job_scripts.app import update as update_job_script
from jobbergate_cli.subapps.job_submissions.app import create as create_job_submission
from jobbergate_cli.subapps.job_submissions.app import delete as delete_job_submission
from jobbergate_cli.subapps.job_submissions.app import get_one as get_job_submission
from jobbergate_cli.subapps.job_submissions.app import list_all as list_job_submissions


def add_legacy_compatible_commands(app: typer.Typer):
    """
    Add commands from the restructured CLI under the previous names for the commands
    to the root ``typer`` app.
    """

    # Applications
    app.command(
        name="list-applications",
        help="LIST the available applications",
    )(list_applications)
    app.command(
        name="get-application",
        help="GET an Application.",
    )(get_application)
    app.command(
        name="create-application",
        help="CREATE an Application.",
    )(create_application)
    app.command(
        name="delete-application",
        help="DELETE an Application.",
    )(delete_application)
    app.command(
        name="update-application",
        help="UPDATE an Application.",
    )(update_application)
    app.command(
        name="download-application",
        help="Download application files.",
    )(download_files_application)
    app.command(
        name="clone-application",
        help="Clone an application.",
    )(clone_application)

    # Job Scripts
    app.command(
        name="list-job-scripts",
        help="LIST job scripts",
    )(list_job_scripts)
    app.command(
        name="get-job-script",
        help="GET a job script",
    )(get_job_script)
    app.command(
        name="create-job-script",
        help="CREATE a job script",
    )(render_job_script)
    app.command(
        name="update-job-script",
        help="UPDATE a job script",
    )(update_job_script)
    app.command(
        name="render-job-script-locally",
        help="Render a job script locally",
    )(render_job_script_locally)
    app.command(
        name="delete-job-script",
        help="DELETE a job script",
    )(delete_job_script)
    app.command(
        name="download-job-script",
        help="Download job script files.",
    )(download_files_job_script)
    app.command(
        name="show-job-script-files",
        help="Show job script files.",
    )(show_files)
    app.command(
        name="clone-job-script",
        help="Clone a job script.",
    )(clone_job_script)

    # Job Submissions
    app.command(
        name="list-job-submissions",
        help="LIST job submissions",
    )(list_job_submissions)
    app.command(
        name="get-job-submission",
        help="GET a job submission",
    )(get_job_submission)
    app.command(
        name="create-job-submission",
        help="CREATE a job submission",
    )(create_job_submission)
    app.command(
        name="delete-job-submission",
        help="DELETE a job submission",
    )(delete_job_submission)



from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Type, Union, cast

import inquirer
import pydantic

from jobbergate_cli.constants import PaginationChoices
from jobbergate_cli.render import StyleMapper, render_paginated_list_results
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import JobbergateContext, ListResponseEnvelope


def handle_pagination(
    jg_ctx: JobbergateContext,
    url_path: str,
    abort_message: str = "There was an error communicating with the API",
    params: Optional[Dict[str, Any]] = None,
    title: str = "Results List",
    style_mapper: StyleMapper | None = None,
    hidden_fields: List[str] | None = None,
    nested_response_model_cls: Type[pydantic.BaseModel] | None = None,
    value_mappers: Optional[Dict[str, Callable[[Any], Any]]] = None,
):
    assert jg_ctx is not None
    assert jg_ctx.client is not None

    current_page = 1

    while True:
        if params is None:
            params = {}
        params["page"] = current_page

        envelope = cast(
            Union[ListResponseEnvelope[Dict[str, Any]], ListResponseEnvelope[pydantic.BaseModel]],
            make_request(
                jg_ctx.client,
                url_path,
                "GET",
                expected_status=200,
                abort_message=abort_message,
                support=True,
                response_model_cls=(
                    ListResponseEnvelope[dict]
                    if nested_response_model_cls is None
                    # mypy doesn't accept dynamic creation of a type in this way
                    # but pydantic requires this to unpack the results correctly
                    else ListResponseEnvelope[nested_response_model_cls]  # type: ignore
                ),
                params=params,
            ),
        )

        render_paginated_list_results(
            jg_ctx,
            envelope,
            title=title,
            style_mapper=style_mapper,
            hidden_fields=hidden_fields,
            value_mappers=value_mappers,
        )

        if envelope.pages <= 1:
            return

        current_page = envelope.page

        message = "Which page would you like to view?"
        choices = [PaginationChoices.PREVIOUS_PAGE, PaginationChoices.NEXT_PAGE, PaginationChoices.EXIT]

        if current_page == 1:
            answer = inquirer.prompt(
                [
                    inquirer.List(
                        "navigation",
                        message=message,
                        choices=choices[1:],  # remove previous page option
                        default=PaginationChoices.NEXT_PAGE,
                    )
                ]
            )
        elif current_page == envelope.pages:
            answer = inquirer.prompt(
                [
                    inquirer.List(
                        "navigation",
                        message=message,
                        choices=choices[::2],  # remove next page option
                        default=PaginationChoices.EXIT,
                    )
                ]
            )
        else:
            answer = inquirer.prompt(
                [
                    inquirer.List(
                        "navigation",
                        message=message,
                        choices=choices,
                        default=PaginationChoices.NEXT_PAGE,
                    )
                ]
            )

        if not answer:
            return

        if answer["navigation"] == PaginationChoices.NEXT_PAGE:
            current_page += 1
        elif answer["navigation"] == PaginationChoices.PREVIOUS_PAGE:
            current_page -= 1
        else:
            return



"""
Provide tool functions for working with Application data.
"""

import contextlib
import copy
import io
import json
import pathlib
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import yaml
from loguru import logger

from jobbergate_cli.constants import (
    JOBBERGATE_APPLICATION_CONFIG,
    JOBBERGATE_APPLICATION_CONFIG_FILE_NAME,
    JOBBERGATE_APPLICATION_MODULE_FILE_NAME,
    JOBBERGATE_APPLICATION_SUPPORTED_FILES,
    FileType,
)
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.render import render_dict, terminal_message
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import (
    ApplicationResponse,
    JobbergateApplicationConfig,
    JobbergateConfig,
    JobbergateContext,
    LocalApplication,
    LocalTemplateFile,
    LocalWorkflowFile,
)
from jobbergate_cli.subapps.applications.application_base import JobbergateApplicationBase
from jobbergate_cli.subapps.applications.questions import inquirer


def load_default_config() -> Dict[str, Any]:
    """
    Load the default config for an application.
    """
    return copy.deepcopy(JOBBERGATE_APPLICATION_CONFIG)


def fetch_application_data_locally(
    application_path: pathlib.Path,
) -> LocalApplication:
    """
    Retrieve an application from a local directory.

    :param: application_path: The directory containing the application files
    :returns: A LocalApplication instance containing the application data
    """
    Abort.require_condition(application_path.is_dir(), f"Application directory {application_path} does not exist")

    config_file_path = application_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
    Abort.require_condition(
        config_file_path.is_file(), f"Application config file {JOBBERGATE_APPLICATION_CONFIG_FILE_NAME} does not exist"
    )

    module_file_path = application_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
    Abort.require_condition(
        module_file_path.is_file(), f"Application module file {JOBBERGATE_APPLICATION_MODULE_FILE_NAME} does not exist"
    )

    template_files_set = set(application_path.rglob("*.j2")) | set(application_path.rglob("*.jinja2"))
    Abort.require_condition(template_files_set, f"No template files found in {application_path}")

    application_config = load_application_config_from_source(config_file_path.read_text())

    supporting_files = application_config.jobbergate_config.supporting_files or []

    template_files = []

    for complete_template_path in template_files_set:
        relative_template_path = complete_template_path.relative_to(application_path)

        if relative_template_path.as_posix() in supporting_files:
            file_type = FileType.SUPPORT
        else:
            file_type = FileType.ENTRYPOINT

        template_files.append(
            LocalTemplateFile(
                filename=relative_template_path.name,
                path=complete_template_path,
                file_type=file_type,
            )
        )

    workflow_file = LocalWorkflowFile(
        filename=module_file_path.name,
        path=module_file_path,
        runtime_config=application_config.jobbergate_config.dict(),
    )

    return LocalApplication(
        template_vars=application_config.application_config,
        template_files=template_files,
        workflow_files=[workflow_file],
    )


def fetch_application_data(
    jg_ctx: JobbergateContext,
    id: Optional[int] = None,
    identifier: Optional[str] = None,
) -> ApplicationResponse:
    """
    Retrieve an application from the API by ``id`` or ``identifier``.

    :param: jg_ctx:     The JobbergateContext. Needed to access the Httpx client with which to make API calls
    :param: id:         The id of the application to fetch
    :param: identifier: If supplied, look for an application instance with the provided identifier
    :returns: An instance of ApplicationResponse containing the application data
    """
    identification: Any = id
    if id is None and identifier is None:
        raise Abort(
            """
            You must supply either [yellow]id[/yellow] or [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
            warn_only=True,
        )
    elif id is not None and identifier is not None:
        raise Abort(
            """
            You may not supply both [yellow]id[/yellow] and [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
            warn_only=True,
        )
    elif identifier is not None:
        identification = identifier

    # Make static type checkers happy
    assert jg_ctx.client is not None

    stub = f"id={id}" if id is not None else f"identifier={identifier}"
    return cast(
        ApplicationResponse,
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-script-templates/{identification}",
            "GET",
            expected_status=200,
            abort_message=f"Couldn't retrieve application {stub} from API",
            response_model_cls=ApplicationResponse,
            support=True,
        ),
    )


def load_application_data(
    app_data: Union[ApplicationResponse, LocalApplication],
    application_source_file: str,
) -> Tuple[JobbergateApplicationConfig, JobbergateApplicationBase]:
    """
    Validates and loads the data for an application returned from the API's applications GET endpoint.

    As part of the Jobbergate data restructure, sections of the legacy jobbergate.yaml
    are now stored in different tables in the backend. This function reconstructs
    them from app_data.workflow_file.runtime_config and app_data.template_vars
    for backward compatibility.

    :param: app_data: A dictionary containing the application data
    :returns: A tuple containing the application config and the application module
    """
    if not app_data.workflow_files:  # make type checker happy
        raise Abort(
            "No workflow file found in application data",
            subject="Invalid application data",
            log_message="No workflow file found in application data",
        )
    try:
        app_config = JobbergateApplicationConfig(
            jobbergate_config=JobbergateConfig(**app_data.workflow_files[0].runtime_config),
            application_config=app_data.template_vars,
        )
    except Exception as err:
        logger.error("ERR: ", err)
        raise Abort(
            "The application config fetched from the API is not valid",
            subject="Invalid application config",
            support=True,
            log_message="Invalid application config",
            original_error=err,
        )

    try:
        app_module = load_application_from_source(application_source_file, app_config)
    except Exception as err:
        raise Abort(
            "The application source fetched from the API is not valid",
            subject="Invalid application module",
            support=True,
            log_message="Invalid application module",
            original_error=err,
        )

    return (app_config, app_module)


@contextlib.contextmanager
def get_upload_files(application_path: pathlib.Path):
    """
    Context manager to build the ``files`` parameter.

    Open the supplied file(s) and build a ``files`` param appropriate for using
    multi-part file uploads with the client.
    """
    Abort.require_condition(application_path.is_dir(), f"Application directory {application_path} does not exist")

    with contextlib.ExitStack() as stack:
        yield [
            (
                "upload_files",
                (path.name, stack.enter_context(io.open(path, mode="r", newline="")), "text/plain"),
            )
            for path in application_path.rglob("*")
            if path.is_file() and path.suffix in JOBBERGATE_APPLICATION_SUPPORTED_FILES
        ]


def upload_application(
    jg_ctx: JobbergateContext,
    application_path: pathlib.Path,
    application_id: Optional[int],
    application_identifier: Optional[str],
):
    """
    Upload an application given an application path and the application id.

    :param: jg_ctx:                 The JobbergateContext. Needed to access the Httpx client with which to
    make API calls
    :param: application_path:       The directory where the application files to upload may be found
    :param: application_id:         The id of the application for which to upload  data
    :param: application_identifier: The identifier of the application for which to upload  data
    """

    # Make static type checkers happy
    assert jg_ctx.client is not None

    identification: Any = application_id
    if application_id is None and application_identifier is None:
        raise Abort(
            """
            You must supply either [yellow]id[/yellow] or [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
            warn_only=True,
        )
    elif application_id is not None and application_identifier is not None:
        raise Abort(
            """
            You may not supply both [yellow]id[/yellow] and [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
            warn_only=True,
        )
    elif application_identifier is not None:
        identification = application_identifier

    Abort.require_condition(application_path.is_dir(), f"Application directory {application_path} does not exist")

    config_file_path = application_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
    Abort.require_condition(config_file_path.is_file(), f"Application config file {config_file_path} does not exist")

    module_file_path = application_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
    Abort.require_condition(module_file_path.is_file(), f"Application module file {module_file_path} does not exist")

    template_files_set = set(application_path.rglob("*.j2")) | set(application_path.rglob("*.jinja2"))
    Abort.require_condition(template_files_set, f"No template files found in {application_path}")

    application_config = load_application_config_from_source(config_file_path.read_text())

    logger.debug("Preparing to upload the template configuration")

    make_request(
        jg_ctx.client,
        f"/jobbergate/job-script-templates/{identification}",
        "PUT",
        expect_response=False,
        abort_message="Request to upload application configuration was not accepted by the API",
        support=True,
        json={"template_vars": application_config.application_config},
        expected_status=200,
    )

    supporting_files = application_config.jobbergate_config.supporting_files or []

    for complete_template_path in template_files_set:
        relative_template_path = complete_template_path.relative_to(application_path)
        logger.debug(f"Preparing to upload {relative_template_path}")

        with open(complete_template_path, "rb") as template_file:
            if relative_template_path.as_posix() in supporting_files:
                file_type = FileType.SUPPORT
            else:
                file_type = FileType.ENTRYPOINT

            make_request(
                jg_ctx.client,
                f"/jobbergate/job-script-templates/{identification}/upload/template/{file_type.value}",
                "PUT",
                expect_response=False,
                abort_message="Request to upload application files was not accepted by the API",
                support=True,
                files={"upload_file": (relative_template_path.name, template_file, "text/plain")},
                expected_status=200,
            )

    logger.debug(f"Preparing to upload {JOBBERGATE_APPLICATION_MODULE_FILE_NAME}")
    with open(module_file_path, "rb") as module_file:
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-script-templates/{identification}/upload/workflow",
            "PUT",
            expect_response=False,
            abort_message="Request to upload application module was not accepted by the API",
            support=True,
            files={
                "upload_file": (module_file_path.name, module_file, "text/plain"),
            },
            data={"runtime_config": application_config.jobbergate_config.json()},
            expected_status=200,
        )


def save_application_files(
    jg_ctx: JobbergateContext,
    application_data: ApplicationResponse,
    destination_path: pathlib.Path,
) -> List[pathlib.Path]:
    """
    Save the application files from the API response into a local destination.
    """
    # Make static type checkers happy
    assert jg_ctx.client is not None

    logger.debug(f"Saving application files to {destination_path.as_posix()}")
    saved_files: List[pathlib.Path] = []

    if application_data.workflow_files:
        application_config = JobbergateApplicationConfig(
            application_config=application_data.template_vars,
            jobbergate_config=JobbergateConfig(**application_data.workflow_files[0].runtime_config),
        )
        config_path = destination_path / JOBBERGATE_APPLICATION_CONFIG_FILE_NAME
        config_path.parent.mkdir(parents=True, exist_ok=True)
        config_path.write_text(
            yaml.dump(
                json.loads(  # workaround to convert path to string
                    application_config.json(
                        exclude_none=True,
                        exclude_unset=True,
                    ),
                ),
                indent=2,
            )
        )
        saved_files.append(config_path)

    for template_file in application_data.template_files:
        template_path = destination_path / "templates" / template_file.filename
        make_request(
            jg_ctx.client,
            template_file.path,
            "GET",
            expected_status=200,
            abort_message=f"Couldn't retrieve template file {template_file.filename} from API",
            save_to_file=template_path,
        )
        saved_files.append(template_path)

    if application_data.workflow_files:
        workflow_path = destination_path / JOBBERGATE_APPLICATION_MODULE_FILE_NAME
        make_request(
            jg_ctx.client,
            application_data.workflow_files[0].path,
            "GET",
            expected_status=200,
            abort_message="Couldn't retrieve application module file from API",
            save_to_file=workflow_path,
        )
        saved_files.append(workflow_path)

    logger.debug(f"The following files were saved: {list(map(str, saved_files))}")
    return saved_files


def load_application_config_from_source(config_source: str) -> JobbergateApplicationConfig:
    """
    Load the JobbergateApplicationConfig from a text string containing the config as YAML.

    :param: config_source: The YAML containing the config
    :returns: A JobbergateApplicationConfig instance with the config values
    """
    config_data = yaml.safe_load(config_source)
    config = JobbergateApplicationConfig(**config_data)
    return config


def load_application_from_source(app_source: str, app_config: JobbergateApplicationConfig) -> JobbergateApplicationBase:
    """
    Load the JobbergateApplication class from a text string containing the source file.

    Creates the module in a temporary file and imports it with importlib.

    Adapted from: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly

    :param: app_source: The JobbergateApplication source code to load
    :param: app_config: The JobbergateApplicationConfig needed to instantiate the JobbergateApplication
    """
    app_locals: Dict[str, Any] = dict()
    exec(app_source, app_locals, app_locals)
    jobbergate_application_class = app_locals["JobbergateApplication"]
    application = jobbergate_application_class(app_config.dict())
    return application


@dataclass
class ApplicationRuntime:
    """
    Prepare and execute a Jobbergate application gathering the answers to the questions.

    :param app_data: The application data, can be either an ApplicationResponse or a LocalApplication.
    :param app_source_code: The source code of the application, often coming from jobbergate.py file.
    :param supplied_params: The parameters supplied to the application, defaults to an empty dictionary.
    :param fast_mode: A flag indicating whether the application is in fast mode, defaults to False.
    """

    app_data: Union[ApplicationResponse, LocalApplication]
    app_source_code: str
    supplied_params: Dict[str, Any] = field(default_factory=dict)
    fast_mode: bool = False

    def __post_init__(self):
        self.app_config, self.app_module = load_application_data(self.app_data, self.app_source_code)
        self.answers: Dict[str, Any] = dict()

    def execute_application(self):
        """Execute the jobbergate application python module."""
        try:
            self._gather_answers()
        except Exception as err:
            exception_name = type(err).__name__
            terminal_message(
                "The question workflow failed to execute. Please check the traceback bellow for more information.",
                subject=f"Runtime error on application execution - {exception_name}",
                color="red",
            )
            raise err

        self._update_template_files_information()

    def as_flatten_param_dict(self) -> Dict[str, Any]:
        """Flatten the internal data to support the rendering process."""
        param_dict_flat = {}
        for outer_value in self.app_config.dict().values():
            for nest_key, nest_value in outer_value.items():
                param_dict_flat[nest_key] = nest_value
        return param_dict_flat

    def _gather_answers(self):
        """Gather the parameter values by executing the application methods."""
        logger.debug("Gathering answers from the application")
        self.answers.update(self.supplied_params)
        # config should be self.answers ideally
        # but we use self.app_module.jobbergate_config for backward compatibility with v1
        config = self.app_module.jobbergate_config
        config.update(self.supplied_params)

        next_method = "mainflow"

        while next_method is not None:
            method_to_call = getattr(self.app_module, next_method)
            logger.debug(f"Calling method {next_method}")
            try:
                workflow_questions = method_to_call(data=config)
            except NotImplementedError:
                raise Abort(
                    f"""
                    Abstract method not implemented.

                    Please implement {next_method} in your class.",
                    """,
                    subject="Invalid application module",
                )

            prompts = []
            auto_answers = {}

            if workflow_questions is None:
                logger.warning(
                    "Deprecation warning: Application method {} returned None while a list is expected", next_method
                )
                workflow_questions = []

            for question in workflow_questions:
                if question.variablename in self.supplied_params:
                    continue
                elif self.fast_mode and question.default is not None:
                    auto_answers[question.variablename] = question.default
                else:
                    prompts.extend(question.make_prompts())

            workflow_answers = cast(Dict[str, Any], inquirer.prompt(prompts, raise_keyboard_interrupt=True))
            workflow_answers.update(auto_answers)

            logger.debug(f"Answers gathered from {next_method}: {workflow_answers}")

            config.update(workflow_answers)
            self.answers.update(workflow_answers)
            if len(auto_answers) > 0:
                render_dict(auto_answers, title="Default values used")

            next_method = config.pop("nextworkflow", None)

        # Legacy applications change the values at runtime, so we need to update the config
        self.app_config = JobbergateApplicationConfig(
            application_config=dict(self.app_module.application_config),
            jobbergate_config=JobbergateConfig.parse_obj(self.app_module.jobbergate_config),
        )
        logger.debug(f"Concluded getting answers: {self.answers}")

    def _update_template_files_information(self):
        """Update the information about the template files if not already present in the configuration."""
        if not self.app_config.jobbergate_config.default_template:
            list_of_entrypoints = [
                i.filename for i in self.app_data.template_files if i.file_type.upper() == "ENTRYPOINT"
            ]
            if len(list_of_entrypoints) != 1:
                raise Abort(
                    f"""
                    Application does not have one entry point, found {len(list_of_entrypoints)}",
                    """,
                    subject="Entry point is unspecified",
                    log_message="Entry point file not specified",
                )
            self.app_config.jobbergate_config.default_template = list_of_entrypoints[0]

        if not self.app_config.jobbergate_config.supporting_files:
            list_of_supporting_files = [
                i.filename for i in self.app_data.template_files if i.file_type.upper() == "SUPPORT"
            ]
            if list_of_supporting_files:
                self.app_config.jobbergate_config.supporting_files = list_of_supporting_files
